/* re_module.c - __re module (the C part of the re module)

   Copyright (c) 2010 Jukka Lehtosalo

   Alore is licensed under the terms of the MIT license.  See the file
   LICENSE.txt in the distribution.
*/

#include "alore.h"
#include "re.h"
#include "debug_params.h"


static AValue BuildMatchObject(AThread *t, AValue *frame, int numGroups,
                              AReRange *subExp);
static ABool GetArguments(AThread *t, AValue *frame, int *pos);


static int MatchResultNum;
static int IgnoreCaseNum;
static int RegExpNum;
int ARegExpErrorNum;


#ifdef A_DEBUG
static AValue ReDump(AThread *t, AValue *frame)
{
    AValue reVal;
    ARegExp *re;
    
    if (!AIsStr(frame[0]))
        return ARaiseTypeErrorND(t);

    reVal = ACompileRegExp(t, frame[0], 0);
    if (AIsError(reVal))
        return reVal;

    re = AValueToPtr(reVal);
    ADisplayRegExp(re);
    
    return ANil;
}
#endif


static AValue ReMatch(AThread *t, AValue *frame)
{
    ARegExp *re;
    AReRange subExp[11];
    int status;
    int pos;

    if (!GetArguments(t, frame, &pos))
        return AError;

    status = AMatchRegExp(frame + 2, frame + 1, AStrLen(frame[1]), pos,
                          A_RE_NOSEARCH, subExp, 10);
    if (status > 0) {
        re = AValueToPtr(frame[2]);
        return BuildMatchObject(t, frame, re->numGroups, subExp);
    } else if (status == 0)
        return ANil;
    else
        return ARaiseMemoryErrorND(t);
}


static AValue ReSearch(AThread *t, AValue *frame)
{
    ARegExp *re;
    AReRange subExp[11];
    int status;
    int pos;

    if (!GetArguments(t, frame, &pos))
        return AError;

    status = AMatchRegExp(frame + 2, frame + 1, AStrLen(frame[1]), pos, 0,
                         subExp, 10);
    if (status > 0) {
        re = AValueToPtr(frame[2]);
        return BuildMatchObject(t, frame, re->numGroups, subExp);
    } else if (status == 0)
        return ANil;
    else
        return ARaiseMemoryErrorND(t);
}


/* Process the arguments to ReMatch or ReSearch. */
static ABool GetArguments(AThread *t, AValue *frame, int *pos)
{
    AExpectStr(t, frame[1]);

    if (!AIsDefault(frame[2])) {
        *pos = AGetInt(t, frame[2]);
        if (*pos < 0 || *pos > AStrLen(frame[1])) {
            ARaiseValueError(t, "Invalid position");
            return FALSE;
        }
    } else
        *pos = 0;

    if (AIsValue(t, frame[0], AGlobalByNum(RegExpNum))) {
        frame[2] = AMemberDirect(frame[0], 0);
        return TRUE;
    }

    AExpectStr(t, frame[0]);
    
    frame[2] = ACompileRegExp(t, frame[0], 0);
    if (AIsError(frame[2]))
        return FALSE;

    return TRUE;
}


static AValue MatchResultGroup(AThread *t, AValue *frame)
{
    AValue a;
    int ind;
    int i1, i2;

    ind = AGetInt(t, frame[1]);
    a = AMemberDirect(frame[0], 0);

    if (ind > (AArrayLen(a) >> 1) - 1)
        return ARaiseValueErrorND(t);
    
    i1 = AValueToInt(AArrayItem(a, 2 * ind));
    i2 = AValueToInt(AArrayItem(a, 2 * ind + 1));

    if (i1 < 0)
        return ANil;
    else
        return ASubStr(t, AMemberDirect(frame[0], 1), i1, i2);
}


static AValue MatchResultSpan(AThread *t, AValue *frame)
{
    AValue a;
    int ind;
    AValue lo;

    ind = AGetInt(t, frame[1]);
    a = AMemberDirect(frame[0], 0);

    if (ind > (AArrayLen(a) >> 1) - 1)
        return ARaiseValueErrorND(t);

    lo = AArrayItem(a, 2 * ind);
    if ((ASignedValue)lo < 0)
        return ANil;
        
    return AMakePair(t, lo, AArrayItem(a, 2 * ind + 1));
}


/* Frame points to the frame of ReMatch or ReSearch. */
static AValue BuildMatchObject(AThread *t, AValue *frame,
                              int numGroups, AReRange *subExp)
{
    AValue match;
    int i;

    /* Create an array containing the locations of the matched groups. */
    frame[3] = AMakeArray(t, numGroups * 2);
    for (i = 0; i < numGroups; i++) {
        int beg = subExp[i].beg;
        int end = subExp[i].end;

        ASetArrayItem(t, frame[3], 2 * i, AIntToValue(beg));
        ASetArrayItem(t, frame[3], 2 * i + 1, AIntToValue(end));
    }

    /* Create the MatchResult object. */
    match = ACallValue(t, AGlobalByNum(MatchResultNum), 0, frame + 4);
    if (AIsError(match))
        return AError;

    frame[4] = match;
    ASetMemberDirect(t, frame[4], 0, frame[3]);
    ASetMemberDirect(t, frame[4], 1, frame[1]);

    return frame[4];
}


static AValue RegExpCreate(AThread *t, AValue *frame)
{
    int flags = 0;
    
    AExpectStr(t, frame[1]);

    if (!AIsDefault(frame[2])) {
        if (frame[2] == AGlobalByNum(IgnoreCaseNum))
            flags |= A_RE_NOCASE;
        else {
            ARaiseValueErrorND(t);
            return AError;
        }
    }

    frame[2] = ACompileRegExp(t, frame[1], flags);
    if (AIsError(frame[2]))
        return FALSE;

    ASetMemberDirect(t, frame[0], 0, frame[2]);

    return frame[0];
}


static AValue ReMain(AThread *t, AValue *frame)
{
    ARegExpErrorNum = AGetGlobalNum(t, "re::RegExpError");
    return ANil;
}


A_MODULE(__re, "__re")
    A_IMPORT("re")

    A_SUB(A_PRIVATE("Main"), 0, 0, ReMain)
    
    A_SUB_OPT("Match", 2, 3, 2, ReMatch)
    A_SUB_OPT("Search", 2, 3, 2, ReSearch)

    A_SYMBOLIC_CONST_P("IgnoreCase", &IgnoreCaseNum)

    A_CLASS_PRIV_P("RegExp", 1, &RegExpNum)
      A_METHOD_OPT("create", 1, 2, 0, RegExpCreate)
    A_END_CLASS()

    A_CLASS_PRIV_P("MatchResult", 2, &MatchResultNum)
        /* FIX: Constructor? The type object is visible via Type() */
        A_METHOD("group", 1, 0, MatchResultGroup)
        A_METHOD("span", 1, 0, MatchResultSpan)
    A_END_CLASS()

#ifdef A_DEBUG    
    A_SUB("__Dump", 1, 0, ReDump)
#endif
A_END_MODULE()
