/* std_int.c - std::Int related operations (see also std_int_long.c)

   Copyright (c) 2010 Jukka Lehtosalo

   Alore is licensed under the terms of the MIT license.  See the file
   LICENSE.txt in the distribution.
*/

#include <math.h>
#include "alore.h"
#include "runtime.h"
#include "memberid.h"
#include "int.h"
#include "str.h"
#include "mem.h"
#include "internal.h"


static AValue RadixInt(AThread *t, AValue *frame);


AValue AStdInt(AThread *t, AValue *frame)
{
    int ind;
    int end;
    unsigned char *buf;
    ABool isNeg;
    ASignedValue num;

    if (!AIsDefault(frame[1]))
        return RadixInt(t, frame);

    if (AIsNarrowStr(frame[0])) {

        ind = 0;
        buf = AValueToStr(frame[0])->elem;
        end = AGetStrLen(AValueToStr(frame[0]));

      TryConvert:

        while (ind < end && (buf[ind] == ' ' || buf[ind] == '\t'))
            ind++;

        isNeg = FALSE;
        if (ind < end) {
            if (buf[ind] == '-') {
                isNeg = TRUE;
                ind++;
            } else if (buf[ind] == '+')
                ind++;
        }

        num = 0;
        
        if (ind < end && AIsDigit(buf[ind])) {
            do {
                int digit = buf[ind] - '0';

                if (num >= (A_SHORT_INT_MAX + 1) / 10
                    && (num > (A_SHORT_INT_MAX + 1) / 10
                        || digit > ((A_SHORT_INT_MAX + 1) -
                                    ((A_SHORT_INT_MAX + 1) / 10) * 10)))
                    goto HandleLongInt;

                num = num * 10 + digit;

                ind++;
            } while (ind < end && AIsDigit(buf[ind]));
        } else
            return ARaiseValueErrorND(t);

        for (; ind < end; ind++) {
            if (buf[ind] != ' ' && buf[ind] != '\t')
                return ARaiseValueErrorND(t);
        }

        if (isNeg)
            return AIntToValue(-num);
        else if (num == A_SHORT_INT_MAX + 1)
            return ACreateLongIntFromIntND(t, num);
        else
            return AIntToValue(num);
    } else if (AIsSubStr(frame[0])) {
        ASubString *ss = AValueToSubStr(frame[0]);

        ind = AValueToInt(ss->ind);
        end = ind + AValueToInt(ss->len);

        if (AIsNarrowStr(ss->str)) {
            frame[0] = ss->str;
            buf = AValueToStr(frame[0])->elem;
            goto TryConvert;
        } else {
            frame[0] = ss->str;
            goto Wide;
        }
    } else if (AIsFloat(frame[0])) {
        double f = AValueToFloat(frame[0]);
        if (f < A_SHORT_INT_MAX + 1 && f >= A_SHORT_INT_MIN)
            return AIntToValue((ASignedValue)f);
        else
            return AFloatToLongInt(t, f);
    } else if (AIsInstance(frame[0])) {
        AValue v = ACallMethodByNum(t, AM__INT, 0, frame);
        /* Expect an Int object. */
        if (AIsShortInt(v) || AIsLongInt(v))
            return v;
        else if (AIsError(v))
            return v;
        else
            return ARaiseTypeErrorND(t);
    } else if (AIsWideStr(frame[0])) {
        AValue n;

        ind = 0;
        end = AGetWideStrLen(AValueToWideStr(frame[0]));
        
      Wide:

        n = ASubStr(t, frame[0], ind, end);
        
        end -= ind;
        ind = 0;

        frame[0] = n;
        n = AWideStringToNarrow(t, frame[0], frame);
        if (AIsError(n))
            return AError;

        frame[0] = n;
        buf = AValueToStr(frame[0])->elem;
        
        goto TryConvert;
    } else if (AIsShortInt(frame[0]) || AIsLongInt(frame[0]))
        return frame[0];
    else
        return ARaiseTypeErrorND(t);

  HandleLongInt:
    
    {
        ALongInt *li;
        AValue val;

        val = ACreateLongIntFromIntND(t, num);
        if (AIsError(val))
            return AError;

        li = AValueToLongInt(val);

        do {
            unsigned char ch = AValueToStr(frame[0])->elem[ind];

            if (!AIsDigit(ch))
                break;
            
            li = AMulAddSingle(t, li, 10, ch - '0');
            if (li == NULL)
                return AError;
            
            ind++;
        } while (ind < end);

        if (isNeg)
            ASetLongIntSign(li);
        
        return ALongIntToValue(li);
    }
}


AValue AAddInts(AThread *t, AValue a, AValue b)
{
    if (AIsShortInt(a) && AIsShortInt(b)) {
        AValue sum = a + b;
        if (!AIsAddOverflow(sum, a, b))
            return sum;
        else
            return ACreateLongIntFromIntND(t, AValueToInt(a) +
                                       AValueToInt(b));
    } else if (AIsLongInt(a) && AIsLongInt(b))
        return AAddLongInt(t, a, b);
    else {
        a = ACoerce(t, a, b, &b);
        if (AIsError(a))
            return a;
        return AAddInts(t, a, b);
    }
}


AValue ASubInts(AThread *t, AValue a, AValue b)
{
    if (AIsShortInt(a) && AIsShortInt(b)) {
        AValue dif = a - b;
        if (!AIsSubOverflow(dif, a, b))
            return dif;
        else
            return ACreateLongIntFromIntND(t, AValueToInt(a) -
                                        AValueToInt(b));
    } else if (AIsLongInt(a) && AIsLongInt(b))
        return ASubLongInt(t, a, b);
    else {
        a = ACoerce(t, a, b, &b);
        if (AIsError(a))
            return a;
        return ASubInts(t, a, b);
    }
}


static AValue RadixInt(AThread *t, AValue *frame)
{
    int i, j;
    int radix = AGetInt(t, frame[1]);
    ABool isNeg;
    
    if (radix < 2 || radix > 36)
        return ARaiseValueError(t, "Invalid radix");
    
    frame[1] = AZero;
    
    AExpectStr(t, frame[0]);

    for (i = 0; i < AStrLen(frame[0]); i++) {
        int ch = AStrItem(frame[0], i);
        if (ch != ' ' && ch != '\t')
            break;
    }
    
    if (i < AStrLen(frame[0]) && AStrItem(frame[0], i) == '-') {
        isNeg = TRUE;
        i++;
    } else
        isNeg = FALSE;

    j = i;
    for (; i < AStrLen(frame[0]); i++) {
        int ch = AStrItem(frame[0], i);
        int digit;
        
        if (AIsDigit(ch))
            digit = ch - '0';
        else if (ch >= 'a' && ch <= 'z')
            digit = ch - 'a' + 10;
        else if (ch >= 'A' && ch <= 'Z')
            digit = ch - 'A' + 10;
        else
            break;

        if (digit >= radix)
            ARaiseValueError(t, "Invalid character");

        frame[1] = AMul(t, frame[1], AMakeShortInt(radix));
        frame[1] = AAdd(t, frame[1], AMakeShortInt(digit));
    }

    if (i == j)
        return ARaiseValueError(t, "Argument not number");

    for (; i < AStrLen(frame[0]); i++) {
        int ch = AStrItem(frame[0], i);
        if (ch != ' ' && ch != '\t')
            return ARaiseValueError(t, "Invalid character");
    }

    if (!isNeg)
        return frame[1];
    else
        return ANeg(t, frame[1]);
}


AValue AIntDivMod(AThread *t, ASignedValue left, ASignedValue right,
                AValue *modPtr)
{
    ASignedValue quot;
    ASignedValue mod;

    if (right == AZero)
        return ARaiseArithmeticErrorND(t, AMsgDivisionByZero);
    
    if (left < 0) {
        if (right < 0) {
            if (left == AIntToValue(A_SHORT_INT_MIN)
                    && right == AIntToValue(-1)) {
                *modPtr = AZero;
                return ACreateLongIntFromIntND(t, -A_SHORT_INT_MIN);
            }
            quot = AIntToValue(-AValueToInt(left) / -AValueToInt(right));
        } else
            quot = -AIntToValue(-AValueToInt(left) / AValueToInt(right));
        
    } else {
        if (right < 0)
            quot = -(left / -AValueToInt(right));
        else
            quot = left / AValueToInt(right);
    }

    quot &= ~A_INTEGER_MASK;

    mod = left - quot * AValueToInt(right);
    if ((mod < 0 && right > 0) || (mod > 0 && right < 0)) {
        mod += right;
        quot -= AIntToValue(1);
    }

    *modPtr = mod;
    return quot;
}
