/* std_module.c - std module

   Copyright (c) 2010 Jukka Lehtosalo

   Alore is licensed under the terms of the MIT license.  See the file
   LICENSE.txt in the distribution.
*/

#include <math.h>
#include <stdlib.h>
#include "alore.h"
#include "runtime.h"
#include "memberid.h"
#include "std_module.h"
#include "io_module.h"
#include "mem.h"
#include "array.h"
#include "str.h"
#include "int.h"
#include "float.h"
#include "internal.h"
#include "gc.h"
#include "class.h"
#include "util.h"
#include "wrappers.h"
/* #include "jitcomp.h" */


static AValue ArrayRepr(AThread *t, AValue *array, AValue *str,
                       AValue *tmp, int *indices, int depth);
static AValue PairOrRangeHashValue(AThread *t, AValue *frame);
static AValue ConcatFixArrayAndArray(AThread *t, AValue fixa, AValue a);


static int StdBooleanNum;
static int StdArrayNum;
static int StdMapNum;
int AStdObjectNum;
int AStdStrNum;
int AStdReprNum;
int AStdIntNum;
int AStdTypeNum;
int AStdFloatNum;
int AStdRangeNum;
int AStdFunctionNum;
int AStdPairNum;
int AStdConstantNum;
int AStdHashNum;

int AStdBufNum;


int AErrorClassNum[EX_LAST + 1];
static int TrueNum;
static int FalseNum;

static int TabNum;
static int CRNum;
static int LFNum;
static int NewlineNum;


static int RangeIterNum;

int AStdExceptionNum;
int AAnonFuncClassNum;

static int StrClassNum;
static int IntClassNum;
static int FloatClassNum;
static int RangeClassNum;
static int PairClassNum;
static int ConstantClassNum;
static int FunctionClassNum;
static int TypeClassNum;


ATypeInfo *AArrayClass;
ATypeInfo *AStrClass;
ATypeInfo *AIntClass;
ATypeInfo *AFloatClass;
ATypeInfo *ARangeClass;
ATypeInfo *APairClass;
ATypeInfo *AConstantClass;
ATypeInfo *AFunctionClass;
ATypeInfo *ATypeClass;

ATypeInfo *AAnonFuncClass;


ABool AInitializeStdTypes(void)
{
    AArrayClass = AValueToType(AGlobalByNum(StdArrayNum));
    AStrClass = AValueToType(AGlobalByNum(StrClassNum));
    AIntClass = AValueToType(AGlobalByNum(IntClassNum));
    AFloatClass = AValueToType(AGlobalByNum(FloatClassNum));
    ARangeClass = AValueToType(AGlobalByNum(RangeClassNum));
    APairClass = AValueToType(AGlobalByNum(PairClassNum));
    AConstantClass = AValueToType(AGlobalByNum(ConstantClassNum));
    AFunctionClass = AValueToType(AGlobalByNum(FunctionClassNum));
    ATypeClass = AValueToType(AGlobalByNum(TypeClassNum));
    AAnonFuncClass = AValueToType(AGlobalByNum(AAnonFuncClassNum));
    return TRUE;
}    


ABool AInitializeStdExceptions(AThread *t)
{
    int i;

    /* IDEA: No need to create ExitException instance and maybe some others. */

    /* IDEA: Maybe all of these can be removed, since these are inherently
             thread-unsafe. */
    
    for (i = 0; i <= EX_LAST; i++) {
        AValue obj = ACreateBasicExceptionInstance(t, i, NULL);
        if (!ASetConstGlobalValue(t, GL_FIRST_ERROR_INSTANCE + i, obj))
            return FALSE;
    }

    return TRUE;
}


static AValue ObjectCreate(AThread *t, AValue *frame)
{
    return frame[0];
}


static AValue Object_eq(AThread *t, AValue *frame)
{
    return frame[0] == frame[1] ? ATrue : AFalse;
}


AValue AStdChr(AThread *t, AValue *frame)
{
    ASignedValue i;
    ABool result;
    
    if (!AIsShortInt(frame[0])) {
        if (AIsLongInt(frame[0]))
            return ARaiseValueErrorND(t);
        else
            return ARaiseTypeErrorND(t);
    }

    i = AValueToInt(frame[0]);
    if ((AValue)i < 256) {
        AString *s;

        AAlloc_M(t, AGetBlockSize(sizeof(AValue) + 1), s, result);
        if (!result)
            return ARaiseMemoryErrorND(t);

        AInitNonPointerBlock(&s->header, 1);
        s->elem[0] = i;

        return AStrToValue(s);
    } else if ((AValue)i < 65536) {
        AWideString *s;

        AAlloc_M(t, AGetBlockSize(sizeof(AValue) + sizeof(AWideChar)), s,
                result);
        if (!result)
            return ARaiseMemoryErrorND(t);

        AInitNonPointerBlock(&s->header, sizeof(AWideChar));
        s->elem[0] = i;

        return AWideStrToValue(s);
    } else
        return ARaiseValueErrorND(t);

    /* Not reached */
}


static AValue StdOrd(AThread *t, AValue *frame)
{
    if (AIsNarrowStr(frame[0])) {
        if (AGetStrLen(AValueToStr(frame[0])) != 1)
            return ARaiseValueErrorND(t);

        return AIntToValue(AValueToStr(frame[0])->elem[0]);
    } else if (AIsWideStr(frame[0])) {
        if (AGetWideStrLen(AValueToWideStr(frame[0])) != 1)
            return ARaiseValueErrorND(t);
        
        return AIntToValue(AValueToWideStr(frame[0])->elem[0]);
    } else
        return ARaiseTypeErrorND(t);
}


static AValue AMakeUnmovableBoolean(AThread *t, ABool bool)
{
    AValue *obj = AAllocUnmovable(2 * sizeof(AValue));
    if (obj == NULL)
        return ARaiseMemoryError(t);
    AInitInstanceBlockGen(obj, AGlobalByNum(StdBooleanNum), 0);
    ((AInstance *)obj)->member[0] = AIntToValue(bool);
    return AInstanceToValue((AInstance *)obj);
}


static AValue StdMain(AThread *t, AValue *frame)
{
    /* Initialize the True object. */
    frame[0] = AMakeUnmovableBoolean(t, TRUE);
    ASetGlobalByNum(GL_TRUE, frame[0]);
    ASetGlobalByNum(TrueNum, frame[0]);
    
    /* Initialize the False object. */
    frame[0] = AMakeUnmovableBoolean(t, FALSE);
    ASetGlobalByNum(GL_FALSE, frame[0]);
    ASetGlobalByNum(FalseNum, frame[0]);

    ATrue = AGlobalByNum(TrueNum);
    AFalse = AGlobalByNum(FalseNum);
    
    ASetConstGlobalByNum(t, TabNum, AMakeStr(t, "\t"));
    ASetConstGlobalByNum(t, CRNum, AMakeStr(t, "\r"));
    ASetConstGlobalByNum(t, LFNum, AMakeStr(t, "\n"));
    ASetConstGlobalByNum(t, NewlineNum, AMakeStr(t, A_NEWLINE_STRING));

#ifdef HAVE_JIT_COMPILER
    {
        AValue map = ACallValue(t, AGlobalByNum(StdMapNum), 0, frame);
        if (AIsError(map))
            return AError;
        ASetGlobalByNum(AJitInfoMapNum, map);
    }
#endif
    
    return AIoMain(t, frame);
}


#define MAX_REPR_CALL_DEPTH 16
#define MAX_REPR_LEN 10000


AValue AStdRepr(AThread *t, AValue *frame)
{
    AValue *tmp;
    AValue retVal;
    
    if (!AAllocTempStack(t, 4))
        return AError;

    tmp = t->tempStackPtr - 4;
    tmp[0] = AZero;
    tmp[1] = AZero;
    tmp[2] = AZero;
    tmp[3] = AZero;
    
    if (AIsShortInt(frame[0]) || AIsLongInt(frame[0]))
        retVal = AStdStr(t, frame);
    else if (AIsStr(frame[0]))
        retVal = AStrRepr(t, frame);
    else if (AIsArray(frame[0])) {
        int indices[MAX_REPR_CALL_DEPTH + 1];
        *tmp = AMakeArray(t, 0);
        retVal = ArrayRepr(t, frame, tmp, tmp + 1, indices, 0);
        if (!AIsError(retVal))
            retVal = AJoin(t, *tmp, ADefault);
    } else if (AIsInstance(frame[0])) {
        if (AMemberByNum(t, frame[0], AM__REPR) == AError) {
            retVal = AStdStr(t, frame);
        } else
            retVal = ACallMethodByNum(t, AM__REPR, 0, frame);
    } else if (AIsConstant(frame[0]) || AIsGlobalFunction(frame[0])
               || AIsNonSpecialType(frame[0]))
        retVal = AStdStr(t, frame);
    else if (AIsPair(frame[0]) || AIsRange(frame[0])) {
        retVal = AError;
        
        tmp[0] = AIsPair(frame[0])
            ? AValueToMixedObject(frame[0])->data.pair.head
            : AValueToMixedObject(frame[0])->data.range.start;
        tmp[0] = ACallValue(t, AGlobalByNum(AStdReprNum), 1, tmp);
        if (tmp[0] != AError) {
            tmp[1] = AIsPair(frame[0])
                ? AValueToMixedObject(frame[0])->data.pair.tail
                : AValueToMixedObject(frame[0])->data.range.stop;
            tmp[1] = ACallValue(t, AGlobalByNum(AStdReprNum), 1,
                                  tmp + 1);
            if (tmp[1] != AError) {
                tmp[0] = AConcatStringAndCStr(t, tmp[0], AIsPair(frame[0])
                                             ? " : " : " to ");
                if (tmp[0] != AError)
                    retVal = AConcatStrings(t, tmp[0], tmp[1]);
            }
        }
    } else if (AIsMethod(frame[0])) {
        char buf[1000];
        AValue f;
        f = AValueToMixedObject(frame[0])->data.boundMethod.method;
        AFormatMessage(buf, 1000, "%F",
                      AValueToFunction(AGlobalByNum(AValueToInt(
                          AValueToMixedObject(
                              frame[0])->data.boundMethod.method))));
        retVal = ACreateStringFromCStr(t, buf);
    } else if (AIsFloat(frame[0])) {
        char buf[100];
        /* IDEA: Verify that the length of the buffer is reasonable. */
        sprintf(buf, "%.17g", AValueToFloat(frame[0]));
        /* If the float is an infinity or a NaN, the result of sprintf varies
           from system to system. Normalize any non-standard values. */
        AFixInfAndNanStrings(buf);
        retVal = ACreateStringFromCStr(t, buf);
    } else
        retVal = ARaiseTypeErrorND(t);

    t->tempStackPtr -= 4;

    return retVal;
}


static AValue RecursiveArrayItem(AValue array, int *indices, int depth)
{
    int i;
    
    for (i = 0; i <= depth; i++)
        array = AArrayItem(array, indices[i]);

    return array;
}


static AValue ArrayRepr(AThread *t, AValue *array, AValue *acc,
                       AValue *tmp, int *indices, int depth)
{
    int i;
    int len;
    AValue str;

    str = AMakeStr(t, "[");
    AAppendArray(t, *acc, str);

    len = AArrayLen(RecursiveArrayItem(*array, indices, depth - 1));
    
    for (i = 0; i < len; i++) {
        indices[depth] = i;
        *tmp = RecursiveArrayItem(*array, indices, depth);
        
        if (AIsArray(*tmp)) {
            if (depth == MAX_REPR_CALL_DEPTH)
                return ARaiseValueErrorND(t);
            if (ArrayRepr(t, array, acc, tmp, indices, depth + 1) ==
                AError)
                return AError;
        } else {
            *tmp = ACallValue(t, AGlobalByNum(AStdReprNum), 1, tmp);
            if (*tmp == AError)
                return *tmp;
            
            AAppendArray(t, *acc, *tmp);
        }
        
        if (i < len - 1) {
            str = AMakeStr(t, ", ");
            AAppendArray(t, *acc, str);
        }
        
        if (AArrayLen(*acc) > MAX_REPR_LEN)
            return ARaiseValueErrorND(t);
    }

    str = AMakeStr(t, "]");
    AAppendArray(t, *acc, str);

    return *acc;
}


/* IDEA: Move this functionality to aloreapi.c or somewhere else, or remove
         it altogether. */
AValue AStdIter(AThread *t, AValue *frame)
{
    if (AIsInstance(frame[0]))
        return ACallMethodByNum(t, AM_ITERATOR, 0, frame);
    else if (AIsRange(frame[0]))
        return ACallValue(t, AGlobalByNum(RangeIterNum), 1, frame);
    else if (AIsStr(frame[0]))
        return ACallValue(t, AGlobalByNum(AStrIterNum), 1, frame);
    else
        return ARaiseTypeErrorND(t);
}


AValue AStdExceptionCreate(AThread *t, AValue *frame)
{
    if (!AIsDefault(frame[1])) {
        if (!AIsStr(frame[1]))
            return ARaiseTypeErrorND(t);
        ASetMemberDirect(t, frame[0], AError_MESSAGE, frame[1]);
    }

    return frame[0];
}


static AValue StdType(AThread *t, AValue *frame)
{
    return ARaiseValueErrorND(t);
}


static AValue StdRange(AThread *t, AValue *frame)
{
    return ARaiseValueErrorND(t);
}


static AValue StdFunction(AThread *t, AValue *frame)
{
    return ARaiseValueErrorND(t);
}


static AValue StdPair(AThread *t, AValue *frame)
{
    return ARaiseValueErrorND(t);
}


static AValue StdConstant(AThread *t, AValue *frame)
{
    /* IDEA: Maybe we should allow creating Constant objects dynamically. */
    return ARaiseValueErrorND(t);
}


AValue AStdHash(AThread *t, AValue *frame)
{
    if (AIsNarrowStr(frame[0]) || AIsSubStr(frame[0]))
        return AStringHashValue(frame[0]);
    if (AIsShortInt(frame[0])) 
        return frame[0] ^ AIntToValue(frame[0] >> 10);
    else if (AIsArray(frame[0]))
        return AArrayHashValue(t, frame);
    else if (AIsInstance(frame[0])) {
        if (AGetInstanceType(AValueToInstance(frame[0]))->hasHashOverload)
            return ACallMethodByNum(t, AM__HASH, 0, frame);
        else
            return AGetIdHashValue(t, frame + 0);
    } else if (AIsWideStr(frame[0]))
        return AStringHashValue(frame[0]);
    else if (AIsConstant(frame[0]))
        return AIntToValue(frame[0] >> 3);
    else if (AIsPair(frame[0]) || AIsRange(frame[0]))
        return PairOrRangeHashValue(t, frame);
    else if (AIsFloat(frame[0]))
        return AFloatHashValue(AValueToFloat(frame[0]));
    else if (AIsLongInt(frame[0]))
        return ALongIntHashValue(frame[0]);
    else if (AIsGlobalFunction(frame[0]) || AIsNonSpecialType(frame[0]))
        return AIntToValue(AValueToInt(frame[0]));
    else if (AIsMethod(frame[0]))
        return AGetIdHashValue(t, frame + 0);
    else
        return ARaiseTypeErrorND(t);
}


static AValue PairOrRangeHashValue(AThread *t, AValue *frame)
{
    AValue h1;
    AValue h2;
    
    frame[1] = AValueToMixedObject(frame[0])->data.range.start;
    h1 = ACallValue(t, AGlobalByNum(AStdHashNum), 1,
                      frame + 1);
    if (AIsError(h1))
        return AError;
    if (!AIsShortInt(h1)) {
        if (AIsLongInt(h1))
            h1 = ALongIntHashValue(h1);
        else
            return ARaiseTypeErrorND(t);
    }
    
    frame[1] = AValueToMixedObject(frame[0])->data.range.stop;
    h2 = ACallValue(t, AGlobalByNum(AStdHashNum), 1,
                      frame + 1);
    if (AIsError(h2))
        return AError;
    if (!AIsShortInt(h2)) {
        if (AIsLongInt(h2))
            h2 = ALongIntHashValue(h2);
        else
            return ARaiseTypeErrorND(t);
    }
    
    return h1 * 33 + h2;
}


static AValue RangeIterCreate(AThread *t, AValue *frame)
{
    AValue lo, hi;
    
    AGetRange(t, frame[1], &lo, &hi);

    AExpectInt(t, lo);
    AExpectInt(t, hi);

    frame[1] = hi;
    ASetMemberDirect(t, frame[0], 0, lo);
    ASetMemberDirect(t, frame[0], 1, frame[1]);
    
    return frame[0];
}


static AValue RangeIterHasNext(AThread *t, AValue *frame)
{
    return AIsLt(t, AMemberDirect(frame[0], 0), AMemberDirect(frame[0], 1)) ?
        ATrue : AFalse;
}


static AValue RangeIterNext(AThread *t, AValue *frame)
{
    AValue next;
    
    frame[1] = AMemberDirect(frame[0], 0);
    next = AAdd(t, frame[1], AMakeInt(t, 1));
    ASetMemberDirect(t, frame[0], 0, next);
    return frame[1];
}


static AValue ExitExceptionCreate(AThread *t, AValue *frame)
{
    /* Check that the second argument is an integer. */
    AGetInt(t, frame[1]);
    ASetMemberDirect(t, frame[0], A_NUM_EXCEPTION_MEMBER_VARS, frame[1]);
    frame[1] = ADefault;
    return AStdExceptionCreate(t, frame);
}


static AValue StdExit(AThread *t, AValue *frame)
{
    AValue exception;
    
    if (AIsDefault(frame[0]))
        frame[0] = AIntToValue(0);

    exception = ACallValue(t,
                           AGlobalByNum(AErrorClassNum[EX_EXIT_EXCEPTION]),
                           1, frame);
    if (AIsError(exception))
        return AError;

    return ARaiseExceptionValue(t, exception);
}


/* Helper function that maps internal type symbols such as "std::__Array" to
   user-visible type symbols such as "std::Array". Other symbols are returned
   unmodified. */
ASymbolInfo *ARealTypeSymbol(ASymbolInfo *sym)
{
    if (sym == AStrClass->sym)
        return AValueToFunction(AGlobalByNum(AStdStrNum))->sym;
    else if (sym == AIntClass->sym)
        return AValueToFunction(AGlobalByNum(AStdIntNum))->sym;
    else if (sym == AFloatClass->sym)
        return AValueToFunction(AGlobalByNum(AStdFloatNum))->sym;
    else if (sym == ARangeClass->sym)
        return AValueToFunction(AGlobalByNum(AStdRangeNum))->sym;
    else if (sym == APairClass->sym)
        return AValueToFunction(AGlobalByNum(AStdPairNum))->sym;
    else if (sym == AConstantClass->sym)
        return AValueToFunction(AGlobalByNum(AStdConstantNum))->sym;
    else if (sym == AFunctionClass->sym)
        return AValueToFunction(AGlobalByNum(AStdFunctionNum))->sym;
    else if (sym == ATypeClass->sym)
        return AValueToFunction(AGlobalByNum(AStdTypeNum))->sym;
    else
        return sym;
}


/* Helper function that maps user-visible type symbols such as "std::Str" to
   internal type symbols such as "std::__Str". Other symbols are returned
   unmodified. */
ASymbolInfo *AInternalTypeSymbol(ASymbolInfo *sym)
{
    if (sym->num == AStdStrNum)
        return AStrClass->sym;
    else if (sym->num == AStdIntNum)
        return AIntClass->sym;
    else if (sym->num == AStdFloatNum)
        return AFloatClass->sym;
    else if (sym->num == AStdRangeNum)
        return ARangeClass->sym;
    else if (sym->num == AStdPairNum)
        return APairClass->sym;
    else if (sym->num == AStdConstantNum)
        return AConstantClass->sym;
    else if (sym->num == AStdFunctionNum)
        return AFunctionClass->sym;
    else if (sym->num == AStdTypeNum)
        return ATypeClass->sym;
    else
        return sym;
}


static AValue StdMin(AThread *t, AValue *frame)
{
    if (!AIsDefault(frame[1])) {
        int ret = AIsLt(t, frame[0], frame[1]);
        if (ret > 0)
            return frame[0];
        else if (ret == 0)
            return frame[1];
        else
            return AError;
    } else {
        int ret;
        
        frame[2] = AIterator(t, frame[0]);
        if (AIsError(frame[2]))
            return AError;

        while ((ret = ANext(t, frame[2], frame + 3)) > 0) {
            if (AIsDefault(frame[1]))
                frame[1] = frame[3];
            else {
                int ret = AIsLt(t, frame[3], frame[1]);
                if (ret > 0)
                    frame[1] = frame[3];
                else if (ret < 0)
                    break;
            }
        }

        if (ret < 0)
            return AError;
        else if (!AIsDefault(frame[1]))
            return frame[1];
        else
            return ARaiseValueError(t, NULL);
    }
}


static AValue StdMax(AThread *t, AValue *frame)
{
    if (!AIsDefault(frame[1])) {
        int ret = AIsGt(t, frame[0], frame[1]);
        if (ret > 0)
            return frame[0];
        else if (ret == 0)
            return frame[1];
        else
            return AError;
    } else {
        int ret;
        
        frame[2] = AIterator(t, frame[0]);
        if (AIsError(frame[2]))
            return AError;

        while ((ret = ANext(t, frame[2], frame + 3)) > 0) {
            if (AIsDefault(frame[1]))
                frame[1] = frame[3];
            else {
                int ret = AIsGt(t, frame[3], frame[1]);
                if (ret > 0)
                    frame[1] = frame[3];
                else if (ret < 0)
                    break;
            }
        }

        if (ret < 0)
            return AError;
        else if (!AIsDefault(frame[1]))
            return frame[1];
        else
            return ARaiseValueError(t, NULL);
    }
}


static AValue StdReverse(AThread *t, AValue *frame)
{
    int i;
    int len;
    
    if (AIsStr(frame[0])) {
        len = AStrLen(frame[0]);
        if (AIsNarrowStr(frame[0]) || AIsNarrowSubStr(frame[0]))
            frame[1] = AMakeEmptyStr(t, len);
        else
            frame[1] = AMakeEmptyStrW(t, len);
        
        for (i = 0; i < len; i++)
            ASetStrItem(frame[1], len - i - 1, AStrItem(frame[0], i));
    } else {
        len = ALen(t, frame[0]);
        if (len < 0)
            return AError;
        frame[1] = AMakeArray(t, len);
        for (i = 0; i < len; i++) {
            AValue item = AGetItemAt(t, frame[0], i);
            if (AIsError(item))
                return AError;
            if (ASetItemAt(t, frame[1], len - i - 1, item) == AError)
                return AError;
        }
    }
    
    return frame[1];
}


static AValue StdAbs(AThread *t, AValue *frame)
{
    if (AIsShortInt(frame[0])) {
        if ((ASignedValue)frame[0] >= 0)
            return frame[0];
        else if (frame[0] != AIntToValue(A_SHORT_INT_MIN))
            return -(ASignedValue)frame[0];
        else
            return AMakeInt64(t, -A_SHORT_INT_MIN);
    } else if (AIsLongInt(frame[0])) {
        if (AGetLongIntSign(AValueToLongInt(frame[0])))
            return ANegateLongInt(t, frame[0]);
        else
            return frame[0];
    } else if (AIsFloat(frame[0]))
        return AMakeFloat(t, fabs(AValueToFloat(frame[0])));
    else
        return ARaiseTypeError(t, NULL);
}


static AValue Write(AThread *t, AValue *frame)
{
    frame[1] = frame[0];
    frame[0] = AGlobalByNum(AStdOutNum);
    return AStreamWrite(t, frame);
}


static AValue WriteLn(AThread *t, AValue *frame)
{
    frame[1] = frame[0];
    frame[0] = AGlobalByNum(AStdOutNum);
    return AStreamWriteLn(t, frame);
}


static AValue ReadLn(AThread *t, AValue *frame)
{
    frame[1] = frame[0];
    frame[0] = AGlobalByNum(AStdInNum);
    return AStreamReadLn(t, frame);
}


static AValue AnonFuncCall(AThread *t, AValue *frame)
{
    int numArgs;
    int numExposed;
    int totalNumArgs;
    AValue func;
    int funcMinArgs, funcMaxArgs;
    
    /* Call the ordinary global function object that is stored in the anonymous
       function object. */

    /* Query some values. */
    numArgs = AArrayLen(frame[1]);
    numExposed = AFixArrayLen(AMemberDirect(frame[0], A_ANON_EXPOSED_VARS));
    totalNumArgs = numArgs + numExposed;
    func = AMemberDirect(frame[0], A_ANON_IMPLEMENTATION_FUNC);
    funcMinArgs = AValueToFunction(func)->minArgs;
    funcMaxArgs = AValueToFunction(func)->maxArgs;
    
    /* Check the number of arguments. The checking performed by
       ACallValueVarArg is not sufficient since it may receive exposed
       variables as additional arguments that invalidate argument counts in
       exception error messages. */
    if (totalNumArgs < funcMinArgs ||
        (!(funcMaxArgs & A_VAR_ARG_FLAG) && totalNumArgs > funcMaxArgs))
        return ARaiseArgumentErrorND(t, func, numArgs, numExposed);
    
    /* The arguments contain first any exposed variable containers needed by
       the anonymous function and then the ordinary arguments passed by the
       caller. */
    frame[1] = ConcatFixArrayAndArray(
        t, AMemberDirect(frame[0], A_ANON_EXPOSED_VARS), frame[1]);
    
    return ACallValueVarArg(t, func, 0, frame + 1);
}


/* Utility function that concatenates a FixArray and an Array object and
   returns an Array object. */
static AValue ConcatFixArrayAndArray(AThread *t, AValue fixa, AValue a)
{
    AValue *temp = AAllocTemps(t, 3);
    int lenFixa, lenA;
    AValue res;
    int i;

    lenFixa = AFixArrayLen(fixa);
    lenA = AArrayLen(a);
    
    temp[0] = fixa;
    temp[1] = a;
    temp[2] = AMakeArray(t, lenFixa + lenA);

    for (i = 0; i < lenFixa; i++)
        ASetArrayItem(t, temp[2], i, AFixArrayItem(temp[0], i));
    for (i = 0; i < lenA; i++)
        ASetArrayItem(t, temp[2], lenFixa + i, AArrayItem(temp[1], i));

    res = temp[2];

    AFreeTemps(t, 3);

    return res;
}


static AValue AnonFuncStr(AThread *t, AValue *frame)
{
    return AMakeStr(t, "anonymous function");
}


static AValue BooleanStr(AThread *t, AValue *frame)
{
    if (AMemberDirect(frame[0], 0) == AZero)
        return AMakeStr(t, "False");
    else
        return AMakeStr(t, "True");
}


A_MODULE(std, "std")
    A_SUB("Main", 0, 5, StdMain)

    /* Internal symbolic constants used by std::Map and std::Set
       implementations*/
    A_SYMBOLIC_CONST_P(A_PRIVATE("Empty"), &AEmptyMarkerNum)
    A_SYMBOLIC_CONST_P(A_PRIVATE("Removed"), &ARemovedMarkerNum)

    A_CLASS_P("Object", &AStdObjectNum)
        A_METHOD("create", 0, 0, ObjectCreate)
        A_METHOD("_eq", 1, 0, Object_eq)
    A_END_CLASS()

    A_CLASS_PRIV_P("Boolean", 1, &StdBooleanNum)
        A_METHOD("create", 0, 0, AHiddenCreate)
        A_METHOD("_str", 0, 0, BooleanStr)
    A_END_CLASS()
    
    A_CLASS_PRIV_P("Array", A_NUM_ARRAY_SLOTS, &StdArrayNum)
        A_METHOD("create", 0, 0, AArrayInitialize)
        A_METHOD("#i", 0, 0, AArrayInitialize)
        A_METHOD("length", 0, 0, AArrayLengthMethod)
        A_METHOD("find", 1, 1, AArrayFind)
        A_METHOD("index", 1, 1, AArrayIndex) /* NOTE: Identical to find */
        A_METHOD("count", 1, 1, AArrayCount)
        A_METHOD("append", 1, 0, AArrayAppend)
        A_METHOD("insertAt", 2, 1, AArrayInsertAt)
        A_METHOD("removeAt", 1, 1, AArrayRemoveAt)
        A_METHOD("iterator", 0, 3, AArrayIter)
        A_METHOD("copy", 0, 0, AArrayCopy)
        A_METHOD("_get", 1, 0, AArray_get)
        A_METHOD("_set", 2, 0, AArray_set)
        A_METHOD("_add", 1, 1, AArray_add)
        A_METHOD("_mul", 1, 1, AArray_mul)
        A_METHOD("_eq", 1, 2, AArray_eq)
        A_METHOD("_lt", 1, 2, AArray_lt)
        A_METHOD("_gt", 1, 2, AArray_gt)
        A_METHOD("_in", 1, 1, AArray_in)
        A_METHOD("_str", 0, 5, AArray_str)
        A_METHOD("_hash", 0, 2, AArray_hash)
    A_END_CLASS()

    A_CLASS_PRIV_P(A_PRIVATE("ArrayIter"), 3, &AArrayIterNum)
        A_METHOD("create", 2, 0, AArrayIterCreate)
        A_METHOD("hasNext", 0, 0, AArrayIterHasNext)
        A_METHOD("next", 0, 0, AArrayIterNext)
    A_END_CLASS()

    A_CLASS_PRIV_P("Map", 4, &StdMapNum)
        A_METHOD_VARARG("create", 0, 0, 6, AMapCreate)
        A_METHOD("#i", 0, 0, AMapInitialize)
        A_METHOD("_get", 1, 3, AMap_get)
        A_METHOD("_set", 2, 3, AMap_set)
        A_METHOD("hasKey", 1, 3, AMapHasKey)
        A_METHOD("length", 0, 0, AMapLength)
        A_METHOD("remove", 1, 3, AMapRemove)
        A_METHOD("keys", 0, 0, AMapKeys)
        A_METHOD("values", 0, 0, AMapValues)
        A_METHOD("_str", 0, 5, AMap_str)
        A_METHOD("iterator", 0, 1, AMapIter)
    A_END_CLASS()
    
    A_CLASS_PRIV_P(A_PRIVATE("MapIter"), 3, &AMapIterNum)
        A_METHOD("create", 1, 0, AMapIterCreate)
        A_METHOD("hasNext", 0, 0, AMapIterHasNext)
        A_METHOD("next", 0, 0, AMapIterNext)
    A_END_CLASS()

    A_CLASS_PRIV_P(A_PRIVATE("__Str"), 1, &StrClassNum)
        A_METHOD("create", 0, 0, AHiddenCreate)
        A_METHOD("length", 0, 0, AStrLengthMethod)
        A_METHOD("iterator", 0, 2, AStrIter)
        A_METHOD("count", 1, 1, AStrCount)
        A_METHOD("upper", 0, 1, AStrUpper)
        A_METHOD("lower", 0, 1, AStrLower)
        A_METHOD("strip", 0, 0, AStrStrip)
        A_METHOD_OPT("find", 1, 2, 2, AStrFind)
        A_METHOD("index", 1, 2, AStrIndex)
        A_METHOD_VARARG("format", 0, 0, 4, AStrFormat)
        A_METHOD("startsWith", 1, 0, AStrStartsWith)
        A_METHOD("endsWith", 1, 0, AStrEndsWith)
        A_METHOD_OPT("replace", 2, 3, 1, AStrReplace)
        A_METHOD_OPT("split", 0, 2, 2, AStrSplit)
        A_METHOD("join", 1, 2, AStrJoin) /* NOTE: Remember AJoin! */
        A_METHOD("_add", 1, 0, AConcatWrapper)
        A_METHOD("_mul", 1, 0, AMulWrapper)
        A_METHOD("_get", 1, 0, AGetItemWrapper)
        A_METHOD("_eq", 1, 0, AEqWrapper)
        A_METHOD("_lt", 1, 0, ALtWrapper)
        A_METHOD("_gt", 1, 0, AGtWrapper)
        A_METHOD("_in", 1, 0, AInWrapper)
        A_METHOD("_hash", 0, 0, AHashWrapper)
        A_METHOD("_repr", 0, 0, AReprWrapper)
        A_METHOD("_int", 0, 0, AIntWrapper)
        A_METHOD("_float", 0, 0, AFloatWrapper)
    A_END_CLASS()

    A_CLASS_PRIV_P(A_PRIVATE("StrIter"), 3, &AStrIterNum)
        A_METHOD("create", 1, 0, AStrIterCreate)
        A_METHOD("hasNext", 0, 0, AStrIterHasNext)
        A_METHOD("next", 0, 0, AStrIterNext)
    A_END_CLASS()

    A_CLASS_PRIV_P(A_PRIVATE("__Int"), 1, &IntClassNum)
        A_METHOD("create", 0, 0, AHiddenCreate)
        A_METHOD("_add", 1, 0, AAddWrapper)
        A_METHOD("_sub", 1, 0, ASubWrapper)
        A_METHOD("_neg", 0, 0, ANegWrapper)
        A_METHOD("_mul", 1, 0, AMulWrapper)
        A_METHOD("_div", 1, 0, ADivWrapper)
        A_METHOD("_idiv", 1, 0, AIdivWrapper)
        A_METHOD("_mod", 1, 0, AModWrapper)
        A_METHOD("_pow", 1, 0, APowWrapper)
        A_METHOD("_eq", 1, 0, AEqWrapper)
        A_METHOD("_lt", 1, 0, ALtWrapper)
        A_METHOD("_gt", 1, 0, AGtWrapper)
        A_METHOD("_hash", 0, 0, AHashWrapper)
        A_METHOD("_str", 0, 0, AStrWrapper)
        A_METHOD("_float", 0, 0, AFloatWrapper)
    A_END_CLASS()

    A_CLASS_PRIV_P(A_PRIVATE("__Float"), 1, &FloatClassNum)
        A_METHOD("create", 0, 0, AHiddenCreate)
        A_METHOD("_add", 1, 0, AAddWrapper)
        A_METHOD("_sub", 1, 0, ASubWrapper)
        A_METHOD("_neg", 0, 0, ANegWrapper)
        A_METHOD("_mul", 1, 0, AMulWrapper)
        A_METHOD("_div", 1, 0, ADivWrapper)
        A_METHOD("_idiv", 1, 0, AIdivWrapper)
        A_METHOD("_mod", 1, 0, AModWrapper)
        A_METHOD("_pow", 1, 0, APowWrapper)
        A_METHOD("_eq", 1, 0, AEqWrapper)
        A_METHOD("_lt", 1, 0, ALtWrapper)
        A_METHOD("_gt", 1, 0, AGtWrapper)
        A_METHOD("_hash", 0, 0, AHashWrapper)
        A_METHOD("_str", 0, 0, AStrWrapper)
        A_METHOD("_repr", 0, 0, AReprWrapper)
        A_METHOD("_int", 0, 0, AIntWrapper)
    A_END_CLASS()
    
    A_CLASS_PRIV_P(A_PRIVATE("__Range"), 1, &RangeClassNum)
        A_METHOD("create", 0, 0, AHiddenCreate)
        A_GETTER("start", 0, ARangeStart)
        A_GETTER("stop", 0, ARangeStop)
        A_METHOD("_eq", 1, 0, AEqWrapper)
        A_METHOD("_in", 1, 0, AInWrapper)
        A_METHOD("iterator", 0, 0, AIterWrapper)
        A_METHOD("_hash", 0, 0, AHashWrapper)
        A_METHOD("_str", 0, 0, AStrWrapper)
    A_END_CLASS()
    
    A_CLASS_PRIV_P(A_PRIVATE("__Pair"), 1, &PairClassNum)
        A_METHOD("create", 0, 0, AHiddenCreate)
        A_GETTER("left", 0, APairLeft)
        A_GETTER("right", 0, APairRight)
        A_METHOD("_eq", 1, 0, AEqWrapper)
        A_METHOD("_hash", 0, 0, AHashWrapper)
        A_METHOD("_str", 0, 0, AStrWrapper)
    A_END_CLASS()
    
    A_CLASS_PRIV_P(A_PRIVATE("__Constant"), 1, &ConstantClassNum)
        A_METHOD("create", 0, 0, AHiddenCreate)
        A_METHOD("_str", 0, 0, AStrWrapper)
        A_METHOD("_eq", 1, 0, AEqWrapper)
    A_END_CLASS()
    
    A_CLASS_PRIV_P(A_PRIVATE("__Function"), 1, &FunctionClassNum)
        A_METHOD("create", 0, 0, AHiddenCreate)
        A_METHOD_VARARG("_call", 0, 0, 1, ACallWrapper)
        A_METHOD("_str", 0, 0, AStrWrapper)
    A_END_CLASS()
    
    A_CLASS_PRIV_P(A_PRIVATE("__Type"), 1, &TypeClassNum)
        A_METHOD("create", 0, 0, AHiddenCreate)
        A_METHOD("supertype", 0, 0, ATypeSupertype)
        A_METHOD("_str", 0, 0, AStrWrapper)
    A_END_CLASS()

    /* Internal anonymous function class (user-visible type is Function) */
    A_CLASS_PRIV_P(A_PRIVATE("AnonFunc"), A_NUM_ANON_SLOTS, &AAnonFuncClassNum)
        /* The first slot is a FixArray object of exposed variable objects */
        /* The second slot is a function object that implements the function */
        A_METHOD_VARARG("_call", 0, 0, 0, AnonFuncCall)
        A_METHOD("_str", 0, 0, AnonFuncStr)
    A_END_CLASS()

    A_SUB_P("Str", 1, 0, AStdStr, &AStdStrNum) /* NOTE: don't change args */
    A_SUB_P("Repr", 1, 0, AStdRepr, &AStdReprNum) /* NOTE: don't change args */
    A_SUB_OPT_P("Int", 1, 2, 0, AStdInt, &AStdIntNum)
    A_SUB_P("Type", 0, 0, StdType, &AStdTypeNum)
    A_SUB_P("Float", 1, 0, AStdFloat, &AStdFloatNum)
    A_SUB_P("Range", 0, 0, StdRange, &AStdRangeNum)
    A_SUB_P("Function", 0, 0, StdFunction, &AStdFunctionNum)
    A_SUB_P("Pair", 0, 0, StdPair, &AStdPairNum)
    A_SUB_P("Constant", 0, 0, StdConstant, &AStdConstantNum)
    
    A_SUB("Abs", 1, 0, StdAbs)
    
    A_SUB("Chr", 1, 0, AStdChr)
    A_SUB("Ord", 1, 0, StdOrd)
    
    A_SUB_P("Hash", 1, 2, AStdHash, &AStdHashNum) /* NOTE: don't change */
    A_SUB_OPT("Min", 1, 2, 2, StdMin)
    A_SUB_OPT("Max", 1, 2, 2, StdMax)
    A_SUB_OPT("Sort", 1, 2, 5, AStdSort)
    A_SUB("Reverse", 1, 1, StdReverse)
    
    A_SUB_OPT("Exit", 0, 1, 2, StdExit)

    A_SUB_VARARG("WriteLn", 0, 0, 2, WriteLn)
    A_SUB_VARARG("Write", 0, 0, 2, Write)
    A_SUB("ReadLn", 0, 5, ReadLn)
    
    A_EMPTY_CONST_P("False", &FalseNum)
    A_EMPTY_CONST_P("True", &TrueNum)

    A_EMPTY_CONST_P("Tab", &TabNum)
    A_EMPTY_CONST_P("CR", &CRNum)
    A_EMPTY_CONST_P("LF", &LFNum)
    A_EMPTY_CONST_P("Newline", &NewlineNum)

    A_CLASS_PRIV_P(A_PRIVATE("__RangeIter"), 2, &RangeIterNum)
        A_METHOD("create", 1, 0, RangeIterCreate)
        A_METHOD("hasNext", 0, 0, RangeIterHasNext)
        A_METHOD("next", 0, 1, RangeIterNext)
    A_END_CLASS()
    
    A_CLASS_PRIV_P("Exception", 1, &AStdExceptionNum)
        A_EMPTY_CONST("message")
        A_METHOD_OPT("create", 0, 1, 0, AStdExceptionCreate)
    A_END_CLASS()
    A_CLASS_P("ValueError", AErrorClassNum + EX_VALUE_ERROR)
        A_INHERIT("Exception")
    A_END_CLASS()
    A_CLASS_P("ResourceError", AErrorClassNum + EX_RESOURCE_ERROR)
        A_INHERIT("Exception")
    A_END_CLASS()
    A_CLASS_P("TypeError", AErrorClassNum + EX_TYPE_ERROR)
        A_INHERIT("ValueError")
    A_END_CLASS()
    A_CLASS_P("MemberError", AErrorClassNum + EX_MEMBER_ERROR)
        A_INHERIT("ValueError")
    A_END_CLASS()
    A_CLASS_P("ArithmeticError", AErrorClassNum + EX_ARITHMETIC_ERROR)
        A_INHERIT("ValueError")
    A_END_CLASS()
    A_CLASS_P("IndexError", AErrorClassNum + EX_INDEX_ERROR)
        A_INHERIT("ValueError")
    A_END_CLASS()
    A_CLASS_P("MemoryError", AErrorClassNum + EX_MEMORY_ERROR)
        A_INHERIT("ResourceError")
    A_END_CLASS()
    A_CLASS_P("InterruptException", AErrorClassNum + EX_INTERRUPT_EXCEPTION)
        A_INHERIT("Exception")
    A_END_CLASS()
    A_CLASS_PRIV_P("ExitException", 1, AErrorClassNum + EX_EXIT_EXCEPTION)
        A_INHERIT("Exception")
        A_METHOD("create", 1, 0, ExitExceptionCreate)
    A_END_CLASS()
    A_CLASS_P("RuntimeError", AErrorClassNum + EX_RUNTIME_ERROR)
        A_INHERIT("Exception")
    A_END_CLASS()
    A_CLASS_P("IoError", AErrorClassNum + EX_IO_ERROR)
        A_INHERIT("Exception")
    A_END_CLASS()

    A_CLASS_PRIV_P(A_PRIVATE("__Buf"), 2, &AStdBufNum)
    A_END_CLASS()

#ifdef HAVE_JIT_COMPILER
    A_VAR_P(A_PRIVATE("__JitInfoMap"), &AJitInfoMapNum)
    A_CLASS_P(A_PRIVATE("__JitInfo"), &AJitInfoNum)
        A_VAR("originalFunction")
    A_END_CLASS()
#endif
A_END_MODULE()
